	list p=16f628

;**************************************************************
;*  	Pinbelegung
;*	----------------------------------	
;*	PORTA: 	0 > L-aktiv: Relay fr 0,1  ..  50 MHz Vorverstrker
;*		1 > L-aktiv: Relay fr 0    ..  50 MHz TTL
;*		2 > L-aktiv: Relay fr 0,03 .. 1,3 GHz mit Vorteiler	
;*		3 < L-aktiv: Taster zur Quellenwahl
;*		4 < Takteingang 1:1 / 1:1 / 64:1 bzw 40:1 ---<<<
;*		5 < Typ des HF-Vorteilers: 1=64:1   0=40:1
;*
;*	PORTB:	0 > LCD Display E
;*		1 > LCD Display R/W		
;*		2 > LCD Display RS
;*		3 < Pulslngeneingang ---<<<	
;*		4-7 LCD Display D4 .. D7 
;*
;* Achtung: Belegung von LCD-R/W weicht von  meinem Standard ab
;*	
;**************************************************************
;
;sprut (zero) Bredendiek 12/2004 .. 08/2015
;
; 02/2015 fix des Anzeigefehlers durch uninitialisierten UFlag,7
; Update fr verbesserte Genauigkeit bei sehr niedigen Frequenzen
; Update fr verbesserte Genauigkeit durch hoeheren Takt
;
; 05/2022 fix des Zhlerberlaufs fr 200MHz<f<256MHz (BIN4 -> 40 Bit)	
;	
;	
; 10-stelliger Frequenzzhler mit LCD-Display:
;
; X XXX XXX XXX Hz
;Amp.   XX XXX Hz
;TTL    XX XXX Hz
;64:1   XX XXX Hz
;
;
; X XXX XXX,XXX Hz
;Amp.      <<1 Hz
;TTL       <<1 Hz
;
;
; - XXX XXX,XXX Hz
;TTL       <<1 Hz
;TTL     0,010 Hz
;
;
; - XXX,XXX XXX Hz
;TTL       <<1 Hz
;T  10,000 000 Hz
;
;
;  3 Eingnge
;    -Mode 0: Vorverstrker  
;    -Mode 1: TTL-Pegel
;    -Mode 2: externer Vorteiler 64:1 (U664B) oder 40:1
;
; Prozessor-Takt 20 MHz
;
; Timer0 : Frequenzzaehler / Uhr
; Timer1 : Pulslngenmessung
; Timer2 : 
;
; Mebereiche bezogen auf RA4:
;  1 Hz  ...  4 MHz	Auflsung  2 Hz  PS=8 =  1:1
;  4 MHz ...  8 MHz	Auflsung  4 Hz  PS=0 =  2:1
;  8 MHz ... 16 MHz	Auflsung  8 Hz  PS=1 =  4:1
; 16 MHz ... 32 MHz	Auflsung 16 Hz  PS=2 =  8:1
; 32 MHz ... 64 MHz	Auflsung 32 Hz  PS=3 = 16:1
;
; Mebereiche bezogen auf U664B:
;  1 MHz ... 256 MHz	Auflsung 128 Hz -> gendert auf 256 Hz
;256 MHz ... 512 MHz	Auflsung 256 Hz
;512 MHz ...   1 GHz	Auflsung 512 Hz
;  1 GHz ...   2 GHz	Auflsung 1 kHz
;  2 GHz ...   4 GHz	Auflsung 2 kHz
;
; unterhalb von 8 kHz Umschaltung auf Pulsweitenmessung
;
; Pulsweitenmessung mit 0,1s Auflsung
;
; Prsezisionsmode:
; Zaehlung ueber 2 Sekunden -> 4-fache Genauigkeit
; Zaehlung ueber 4 Sekunden -> 8-fache Genauigkeit
; Pulsweite bis zu 512 Timer1-Zyklen fuer 4 Perioden :Fmin= 0,6 Hz
;
;**********************************************************
; Includedatei fr den 16F628 einbinden

	;#include <P16f628.INC>
	#include "p16f628.inc"

	ERRORLEVEL      -302    	;SUPPRESS BANK SELECTION MESSAGES


; Configuration festlegen: 16F628
; kein MCLRE, Power on Timer, kein Watchdog, HS-Oscillator, kein Brown out, kein LV-programming
	__CONFIG	_MCLRE_OFF & _PWRTE_ON & _WDT_OFF & _HS_OSC & _BODEN_OFF & _LVP_OFF


;**********************************************************
; Variablen festlegen
; fr alle einheitlich
Flags		Equ	0x20
loops		equ	0x21		; timer fr wait
loops2		equ	0x22		; timer fr wait
LcdStatus	equ	0x23		;
LcdDaten	equ	0x24		;
p_copy		equ	0x25
w_copy		Equ	0x26		; Backup fr Akkuregister
s_copy		Equ	0x27		; Backup fr Statusregister


#define		Taste	PORTA,3


; nur fr den Frequenzzhler
BcdDaten	EQU	0x28
BcdDaten1	EQU	0x29
Fehler		EQU	0x2A
; 32-Bit Rechenregister
xw0		EQU	0x2B	;LSB
xw1		EQU	0x2C	; I 
xw2		EQU	0x2D	; I 
xw3		EQU	0x2E	;MSB
;xw4		EQU	0x4D
; Dezimalstellen
HdG		EQU	0x2F	;10 Milliarden
HdHM		EQU	0x30	;9 HundertMillionen
HdZM		EQU	0x31	;8 ZehnMillionen
HdM		EQU	0x32	;7 Millionen
HdHT		EQU	0x33	;6 Hunderttausender
HdZT		EQU	0x34	;5 Zehntausender
HdT		EQU	0x35	;4 Tausender
HdH		EQU	0x36	;3 Hunderter
HdZ		EQU	0x37	;2 Zehner
HdE		EQU	0x38	;1 Einer
HdX		EQU	0x39	; Puffer fr eine Dezimalstelle
; 32-Bit Rechenregister
f0		EQU	0x3A	;LSB
f1		EQU	0x3B	; I 
f2		EQU	0x3C	; I 
f3		EQU	0x3D	;MSB
; 32 Bit Frequenzzhler
BIN4		EQU	0x3E	;LSB
BIN41		EQU	0x3F	; I 
BIN42		EQU	0x40	; I 
BIN43		EQU	0x41	;MSB
BIN44		EQU	0x42	;VMSB :) fr 40-Bit-Berechnung		
; Hilfszellen fr Frequenzmeschleife
HLTest		EQU	0x43	; HL-Flanken-Detektorn fr Timer0-berlauf
UpperCOU	EQU	0x44
HighCOU		EQU	0x45	; high-Teil des Schleifenzhlers
LowCOU		EQU	0x46	; low-Teil des Schleifenzhlers
PreSC		EQU	0x47	; Vorteiler fr die richtige Messung

Mode		EQU	0x48	; Eingangsauswahl
FlipCount	EQU	0x49
MHz		EQU	0x4A
EEWriteData	EQU	0x4B
TastCounter	EQU	0x4C

xw4		EQU	0x4D

; Variablendefinition nur fr Periodenmessung
PeriodeL	EQU	0x4E
PeriodeH	EQU	0x4F
; 32-Bit-Rechenregister
r0		EQU	0x50
r1		EQU	0x51
r2		EQU	0x52
r3		EQU	0x53
; 32-Bit-Rechenregister
g0		EQU	0x54
g1		EQU	0x55
g2		EQU	0x56
g3		EQU	0x57
; 32-Bit-Rechenregister
h0		EQU	0x58
h1		EQU	0x59
h2		EQU	0x5A
h3		EQU	0x5B
; 32-Bit-Rechenregister
d0		EQU	0x5C
d1		EQU	0x5D
d2		EQU	0x5E
d3		EQU	0x5F
; 32-Bit-Rechenregister
i0		EQU	0x60
i1		EQU	0x61
i2		EQU	0x62
i3		EQU	0x63

TMR1LoopL	EQU	0x64
TMR1LoopH	EQU	0x65
StartL		EQU	0x66
StartH		EQU	0x67
PIR1Puffer	EQU	0x68

; letzte Zelle ist 0x7F im 16F628


; Konstanten nur fr den Frequenzzhler
#define		Buttonpressed	Flags, 0		
#define		Inputchanged	Flags, 1

#define		PpmMode		Flags, 2
#define		PpmMode2	Flags, 3

	
; Konstanten nur fr Periodenmessung
#define		Pulsmode	Flags, 4
#define		Unterlauf	Flags, 5	
#define		Pulsmode3	Flags, 6	; 3 Nachkommastellen
#define		Pulsmode6	Flags, 7	; 6 Nachkommastellen


; Variablendefinition nur fr die Uhr
countL		equ	0x4E
countH		equ	0x4F
ES		equ	0x50		; Einer  Sekunden
ZS		equ	0x51		; Zehner Sekunden
EM		equ	0x52		; Einer  Minuten
ZM		equ	0x53		; Zehner Minuten
EH		equ	0x54		; Einer  Stunden
ZH		equ	0x55		; Zehner Stunden
; 54..57	g
; 58..5B	h
; 5C..5F	d
Digit		Equ	0x60		; 8 Sekunden-Zhler
DCFLoopH	equ	0x61
DCFLoopL	equ	0x62
DCF_onoff	equ	0x63
s0		equ	0x64		; startwert
s1		equ	0x65
DCF_pulscount	equ	0x66
DCF_Quality	equ	0x67
DCF_Minuten	equ	0x68
DCFcountL	equ	0x69
DCFcountH	equ	0x6A
UFlags		equ	0x6B


; Konstanten nur fr die Uhr
#define		neueZeit	UFlags,0
#define		DCF_ready	UFlags,1
#define		DCF_ok		UFlags,2
#define		DCF_todo	UFlags,3
#define		DCF_negativ	UFlags,4
#define		DCF_newpulse	UFlags,5
#define		DCF_active	UFlags,6
#define		DCF_dppm	UFlags,7



; Konstanten festlegen fr LCD (Frq & Uhr)
; fr LCD
PortLCo	equ	PORTB		; LCD-Control-Port
PortLDa	equ	PORTB		; LCD-Daten-Port
LcdE	equ	0		; enable Lcd
LcdRw	equ	1		; read Lcd  (nicht write)
LcdRs	equ	2		; Daten Lcd (nicht control)

	
Ini_con Equ	B'00000000'	; TMR0 -> Interupt disable
Ini_opt	Equ	B'00000010'	; pull-up



;**************************************************************
; los gehts mit dem Programm

	org 	0
	goto	Init	

;**************************************************************
; die Interuptserviceroutine fr die Kalibrierungs-Uhr
; und die Periodenmessung

	org 	4 	
intvec
	movwf	w_copy		; w retten
	swapf	STATUS, w 	; STATUS retten
	bcf	STATUS, RP0	; Bank 0
	movwf	s_copy	

	movf	PCLATH, W
	movwf	p_copy
	clrf	PCLATH		; Bank 0 

	; wer lste den Int aus?
	; DCF-Timer1 ueberlauf ? alle 104 ms
	btfsc	PIR1, TMR1IF
	goto	Int_serv_TMR1

	; DCF-Pulse ?		jede Sekunde
	btfsc	PIR1, CCP1IF
	goto	Int_serv_CCP1

	; die Kalibrieruhr ? alle 51 us
	btfsc	INTCON, T0IF
	goto	Int_serv_T0

	clrf	INTCON		; oops, ein problem
	clrf	PIR1		; Int von unbekannt
	goto	Int_end	


; ********* DCF-Timer1 ueberlauf
Int_serv_TMR1
	incf	DCFLoopL, f
	bnz	Int_end_TMR1
	incf	DCFLoopH, f
;	bcf	DCF_ok		; keine DCF-Pulse

Int_end_TMR1
	bcf	PIR1, TMR1IF	; Interupt-Flag lschen
	goto	Int_end




; ********* Intrupt service routine fr DCF77 vom CCP1
; darf hoechstens 50us (250 Befehle) dauern
Int_serv_CCP1
	incf	DCF_pulscount, f
	bsf	DCF_newpulse
	; wieder eine Minute vorbei ?
	btfss	DCF_ready
	goto	Int_end_CCP1

	; Minutenpulse
	; Zeit -> h
	movfw	CCPR1L
	movwf	h0
	movfw	CCPR1H
	movwf	h1
	movfw	DCFLoopL
	movwf	h2
	movfw	DCFLoopH
	movwf	h3
	clrf	DCFLoopL
	clrf	DCFLoopH

	; im Idealfall ist DCF_pulscount = 59d = 0x3B
	incf	DCF_Quality, f
	movfw	DCF_pulscount
	xorlw	0x3B		; = 59 ?
	skpz
	clrf	DCF_Quality
	clrf	DCF_pulscount

	bsf	DCF_todo	; guelige zeit
	goto	Int_end_CCP1

Int_end_CCP1	
	bcf	DCF_ready
	; 0x7270 = 1,5 Sekunden
	movlw	0x72+1
	movwf	DCFcountH	; mu auf 0x+1 stehen, sonst funktioniert Abbruch bei dec nicht
	movlw	0x70
	movwf	DCFcountL

	bcf	PIR1, CCP1IF	; Interupt-Flag lschen
	goto	Int_end

	

; ********* Intrupt service routine fr Uhr mit Timer0
; alle 51,2 us
Int_serv_T0 
	; DCF sind 1,5 Sekunden nach letztem Pulse vergangen ?
	decf	DCFcountL, f
	skpnz
	decf	DCFcountH, f
	skpnz
	bsf	DCF_ready	; ja, es wird nun ein minuten-pulse kommen


	; nun die Uhr
	decf	countL, f
	btfss	STATUS,Z
	goto	Int_end_T0
	decf	countH,f
	btfss	STATUS,Z
	goto	Int_end_T0	; Sekunde noch nicht vorbei

	; eine Sekunde ist vorbei
	decf	Digit,f		; 4-Sekundenzhler verringern
	; 20 MHz
	; 19531.25 Hz-Timer0-Interupt; 20 MHz / 4 / 256 = 19531.25 Hz
	; 4 x 19531.25 Hz = (3 x 19531) + 19532       = 4 Sekunden
	; Digit = 3..1	Teilen durch 19531 = 4C 4B
	; Digit =    0  Teilen durch 19532 = 4C 4C

	; Sekundentimer wieder einstellen auf 19531
	movlw	0x4C+1
	movwf	countH		; mu auf 0x4B+1 stehen, sonst funktioniert Abbruch bei dec nicht
	movlw	0x4B
	movwf	countL
	; bei Digit = 0 mu countL 1x incrementiert werden

	; falls Digit=0 dann 
	;  Digit von 0 auf 4 setzen
	;  Sekundentimer auf 19532 korrigieren
	tstf	Digit		; Digit=0?
	btfss	STATUS,Z
	goto	Int_1
	bsf	Digit,2		; Digit:=4	
	incf	countL,f	; 19531+1 = 19532

Int_1
	bsf	neueZeit

	incf	ES, f		; 1 Sekunden erhhen
	movlw	D'10'
	subwf	ES, w
	btfss	STATUS, Z	
	goto	Int_end_T0	; 1 Sekunden <> 10	

	clrf	ES
	incf	ZS, f		; 10 Sekunden erhhen
	movlw	D'6'
	subwf	ZS, w
	btfss	STATUS, Z	
	goto	Int_end_T0	; 10 Sekunden <> 6	

	clrf	ZS
	incf	EM, f		; 1 Minuten erhhen
	movlw	D'10'
	subwf	EM, w
	btfss	STATUS, Z	
	goto	Int_end_T0	;  1 Minuten <> 10

	clrf	EM
	incf	ZM, f		; 10 Minuten erhhen
	movlw	D'6'
	subwf	ZM, w
	btfss	STATUS, Z	
	goto	Int_end_T0	; 10 Minuten <> 6

	clrf	ZM
	incf	EH, f		; 1 Stunden erhhen
	movlw	D'10'
	subwf	EH, w
	btfss	STATUS, Z	
	goto	Test24		; 1 Stunden <> 10

	clrf	EH
	incf	ZH, f		; 10 Stunden erhhen
	goto	Int_end_T0

Test24
	movlw	D'2'
	subwf	ZH, w
	btfss	STATUS, Z	
	goto	Int_end_T0	; 10 Stunden <> 2
	movlw	D'4'
	subwf	EH, w
	btfss	STATUS, Z	
	goto	Int_end_T0	; 1 Stunden <> 4
	clrf	EH		; Mitternacht
	clrf	ZH
	goto	Int_end_T0

Int_end_T0	
	bcf	INTCON, T0IF	; Interupt-Flag lschen

Int_end
	movf	p_copy, W
	movwf	PCLATH
	swapf	s_copy, w	; STATUS zurck
	movwf	STATUS 
	swapf	w_copy, f	; w zurck mit flags
	swapf	w_copy, w
	retfie
;********************************************************


;********************************************************
; Das Programm beginnt mit der Initialisierung

Init	
	bsf     STATUS, RP0	; Bank 1
	movlw   Ini_opt     	; pull-up on
	movwf   OPTION_REG 
	movlw	B'11111000'	; RA0 .. RA2 outputs, RA3, RA4 input 
	movwf	TRISA		; 
	movlw	B'00001000'	; PortB alle outputs auer RB3
	movwf	TRISB
	bcf     STATUS, RP0	; Bank 0
	movlw	B'00000111'
	movwf	PORTA		; alle Relais aus		
	clrf	PORTB	

	; 16F628 alle Comparatoreingnge auf Digital umschalten
	BSF	CMCON, CM0
	BSF	CMCON, CM1
	BSF	CMCON, CM2
					
	movlw   Ini_con     	; Interupt disable
	movwf   INTCON   
	clrf	UFlags		; garantiert korrekte anzeige

	call	InitLCD		; Display initialisieren

	; Falls Taster gedrckt ist, dann wird die Uhr-Routine gestartet
	btfss	Taste		; Uhr-Routine  starten?
	goto	Uhr

	; Frequenzzhler -Mode vorbereiten
	clrf	Flags
	bsf	Inputchanged	; damit ein Eingang eingeschaltet wird
	clrf	Mode		; Eingang: Vorverstrker
	call	LoadMode
	call	Inputselect	; Relais ansteuern
	call	LcdShowInput	; Signaleingang anzeigen

	; timer0 vorbereiten fr den Frequenzmesser
	bsf     STATUS, RP0	; Bank 1
	bsf	OPTION_REG, T0CS; RA4-Takt zum Timer0
	bsf	OPTION_REG, PSA	; noch ohne Vorteiler
	bcf     STATUS, RP0	; Bank 0

	movlw	1
	movwf	PreSC		; 4:1



;*****************************************************************
;*****************************************************************
;*****************************************************************
;*****************************************************************
; (HighCOU,LowCOU)x20Takte lang wird gezhlt
; das 24-Bit Ergebnis steht in BIN4+2 (MSB) bis BIN4 (LSB)
; virtuell steht davor BIN+3=0
; das mach dann 32Bit
; 
; fr Autorange: 1MHz Resolution bei PT=256:1
; 256 us x 1 MHz / 256 = 1
;
; 10 MHz-Takt
; 256 us =  640 Takte = 32 x 20 Takte , 00,20h=32d	10 MHz
; fr die Messung 500 ms
; 10MHz ; 500 ms = 1250000 Takte = 62500 x 20 Takte, 0,f4,24h=62500d   -Default
;
;
; 20 MHz-Takt
; 256 us = 1280 Takte = 32 x 20 Takte , 00,20h=32d	20 MHz
; fr die Messung 500 ms
; 20MHz ; 500 ms = 2500000 Takte = 125000 x 20 Takte, 0x01 E8 48   -Default
;
Frequ_Loop
	call	LcdZeile2	
	call	CheckTaster		; Modewechsel durch Taster?
	call	Autorange		; grobe Messung (in MHz) -> BIN4
	call	FSetPS			; bestimme prescaler

	; im Mode 2 immer Frequenzzhlermode (64:1)
	btfsc	Mode,1
	goto	Zaehler

	movfw	BIN4
	bnz	Zaehler			; > 1 MHz

	call	FMessung100Hz		; messen mit 100 Hz Auflsung
	movfw	BIN4
	andlw	0xC0
	iorwf	BIN4+1,w
	iorwf	BIN4+2,w
	iorwf	BIN4+3,w
	bnz	Zaehler			; > 6,3 kHz  hier brauche ich eine Hysterese
	goto	Pulslaenge

Zaehler
	; ansonsten weiter zur Frequenzzhlung
	bcf	Pulsmode
	call	LcdShowInput
	call	FLcdSchoTol		; Toleranz anzeigen
	call	FMessung		; nun genaue Messung -> BIN4
	call	FBerechne		; Frequenz:  BIN4 ->f
	call	CheckIf2Low		; falls <8192 dann Pulsmode aktivieren
	call	Display			; f -> 'x xxx xxx xxx'
	goto	Frequ_Loop		; und alles noch mal von vorne


Pulslaenge
	bsf	Pulsmode
	call	PMessung		; -> w und f
	btfss	Pulsmode
	goto	Fail			; Frequenz zu hoch (> 10 kHz))
	btfsc	Buttonpressed
	goto	Break
	call	PBerechnung	 	; frequenz berechnen xw -> f
	btfsc	Buttonpressed
	goto	Break
	call	Display			; f -> 'x xxx xxx xxx'
	call	PLcdSchoTol2		; Toleranz anzeigen
	call	Wait50ms
	btfss	Taste	
	call	Wait50ms
	btfss	Taste	
	call	Wait50ms
	btfss	Taste	
	call	Wait50ms
	goto	Frequ_Loop		; und alles noch mal von vorne

Fail
Break
	clrf	f3
	clrf	f2
	clrf	f1
	clrf	f0
	call	Display
	goto	Frequ_Loop		; und alles noch mal von vorne



;************************************************************************
FBerechne
	clrf    BIN4+4
	; Zhlwert in Frequenz umrechnen
	; Mezeit kompensieren
	btfss	PpmMode		; 2 oder 4 sekunden
	call	Mal2		; 500ms -> 1s

	; nur fr Mode2
	; externen 64:1-Vorteiler kompensieren
	btfss	Mode, 1		; Bei Mode 0 oder 1 berspringen
	goto	No64
	btfsc	PORTA,5
	call	Mal64		;  32-Bit x 64 (BIN4)
	btfss	PORTA,5
	call	Mal40		;  32-Bit x 40 (BIN4)
No64

	; internen Timer0-Vorteiler kompensieren
	btfsc	PreSC,3
	goto	PSNo3		; kein prescaler
	movfw	PreSC
	movwf	Fehler		; sichern fr Berechnung der Frequenz
	incf	Fehler, f
mal
	call	Mal2		; PS kompensieren
	decfsz	Fehler,f
	goto	mal
PSNo3

	; Mezeit kompensieren 1s -> 2s
	btfsc	PpmMode
	call	Durch2
	; Mezeit kompensieren 2s -> 4s
	btfsc	PpmMode2
	call	Durch2

	movfw	BIN4		; BIN4 -> f
	movwf	f0
	movfw	BIN4+1
	movwf	f1
	movfw	BIN4+2
	movwf	f2
	movfw	BIN4+3
	movwf	f3
	return


;************************************************************************
CheckIf2Low
; falls <8192 dann Pulsmode aktivieren
	movfw	f3
	btfss	STATUS,Z	; falls Z=0 dann zhler
	goto	bleibt_Frq
	movfw	f2
	btfss	STATUS,Z	; falls Z=0 dann zhler
	goto	bleibt_Frq
	btfsc	f1,7		; falls Bit=1 dann zhler
	goto	bleibt_Frq
	btfsc	f1,6		; falls Bit=1 dann zhler
	goto	bleibt_Frq
	btfsc	f1,5		; falls Bit=1 dann zhler
	goto	bleibt_Frq
	bsf	Pulsmode
bleibt_Frq
	return


;***********************************************************************************
;***********************************************************************************
;Messung mit optimaler Auflsung (2 Hz x PS)
; 500 ms =   125 000 x 4us = 0x01 E8 48
;   2 s  =   500 000 x 4us = 0x07 A1 20
;   4 s  = 1 000 000 x 4us = 0x0F 42 40
FMessung
	btfsc	PpmMode2
	goto	FMessung4s
	btfsc	PpmMode
	goto	FMessung2s
	; 500 ms
	movlw	0x01+1 		; 
	movwf	UpperCOU 	; 
	movlw	0xE8+1 		; 
	movwf	HighCOU 	; 
	movlw	0x48+1 		; 
	movwf	LowCOU 		;
	movf	PreSC,W 	; PRESC = Vorteiler-Faktor
	call	Messung
	return

FMessung2s
	; 2 sekunden
	movlw	0x07+1 		; 
	movwf	UpperCOU 	; 
	movlw	0xA1+1 		; 
	movwf	HighCOU 	; 
	movlw	0x20+1 		; 
	movwf	LowCOU 		;
	movf	PreSC,W 	; PRESC = Vorteiler-Faktor
	call	Messung
	return

FMessung4s
	; 4 sekunden
	movlw	0x0F+1 		; 
	movwf	UpperCOU 	; 
	movlw	0x42+1 		; 
	movwf	HighCOU 	; 
	movlw	0x40+1 		; 
	movwf	LowCOU 		;
	movf	PreSC,W 	; PRESC = Vorteiler-Faktor
	call	Messung
	return


;***********************************************************************************
; mittlere Messung (in 100Hz)
;PS   2:1 geht fuer < 5 MHz
; 1/50s = 20 ms = 5000 x 4 us = 0x00 13 88
FMessung100Hz
	movlw	0x00+1 		; 
	movwf	UpperCOU 	; 
	movlw	0x13+1 		; 
	movwf	HighCOU 	; 
	movlw	0x88+1 		; 
	movwf	LowCOU 		; 
	movlw	0x00	 	; PRESC = prescaler factor selected 2:1
	call	Messung
	return

;***********************************************************************************
; mittlere Messung (in kHz)
;PS   2:1 geht fuer < 5 MHz
; 1/500s = 2 ms = 500 x 4 us = 0x00 01 F4
FMessungkHz
	movlw	00h+1 		; 
	movwf	UpperCOU 	; 
	movlw	01h+1 		; 
	movwf	HighCOU 	; 
	movlw	0xF4+1 		; 
	movwf	LowCOU 		; 
	movlw	0x00	 	; PRESC = prescaler factor selected 2:1
	call	Messung
	return

;***********************************************************************************
; grobe Messung (in MHz)
;PS   256:1
; 64 x 4 us = 256 us = 0x00 00 40
Autorange
	movlw	00h+1 		; 
	movwf	UpperCOU 	; 
	movlw	00h+1 		; high loop counter for 256s
	movwf	HighCOU 	; HighCOU = hi byte counter
	movlw	40h+1 		; 00040h=.64 cycles=1280,T=256us
	movwf	LowCOU 		; LowCOU = lo byte counter
	movlw	0x07	 	; PRESC = prescaler factor selected
	;call	Messung

;***********************************************************************************
; die eigentliche  Meroutine
; Mezeit = 20 Zyklen x (UpperCOU,HighCOU,LowCOU) = 4 us x (UpperCOU,HighCOU,LowCOU)
; Prescaler in w
Messung
	addlw	20h 		; for PRESC 0,1,2,3... w=20h,21h,22h,23h...; TOSE=0, TOCS=1, PSA=0, PS=0..7
	bsf	STATUS,RP0 	; bank 1
	movwf	OPTION_REG 	; 
	bcf	STATUS,RP0 	; bank 0
	clrf	BIN4+1 		; clear next counter byte
	clrf	BIN4+2 		; clear next counter byte
	clrf	BIN4+3 		; clear next counter byte
	clrf	HLTest	 	; initialize TMR0 overflow detector
	clrf	TMR0 		; initialize TMR0
	goto	MessungStart	; vielleicht besser zu Loop500U springen

Loop500A
	btfss	Taste		; 2     Abbruch der Frequenzzaehlung ?
	goto	MessungAbbruch	; -
Loop500
	nop			; 1
MessungStart
	nop			; 1
Loop500U
	movf	TMR0, W		; 1 1 1 only place where TMR0 is read
	movwf	BIN4		; 1 1 1 rtcc ---> freq0
	rlf	BIN4, W		; 1 1 1 carry <--- TMR0.7
	rlf	HLTest, F	; 1 1 1 rotor <--- carry
	movf	HLTest, W	; 1 1 1 HLTest=TMR0 overflow detect
	andlw	3		; 1 1 1 mask 2 LSbs for edge detect
	xorlw	2		; 1 1 1 00000000 if 1 <--- 0
	btfss	STATUS, Z	; 1 2 2 | skip if TMR0 overflow
	goto	NotOvf1		; 2 - - | if nz
				;       | 5T any case
	incf	BIN4+1, F	; - 1 1 | nsb
NotOvf1				;
	btfsc	STATUS, Z	; 2 2 1 | skip MSB adv ifnot overflow
	incf	BIN4+2, F 	; - - 1 | msb
	nop			; warten
	decfsz	LowCOU, F 	; 1 (2) lo loop counter
	goto	Loop500A 	; 2 (-) 20T total
	decfsz	HighCOU, F 	; (1) hi loop counter
	goto	Loop500 	; (2) 20T total
	decfsz	UpperCOU, F 	; (1) up loop counter
	goto	Loop500U 	; (2) 20T total

	return

MessungAbbruch
	; Abbruch durch Tastendruck
	clrf	BIN4
	clrf	BIN4+1 		; clear next counter byte
	clrf	BIN4+2 		; clear next counter byte
	clrf	BIN4+3 		; clear next counter byte
	

;***********************************************************************************
; Einstellung des Vorteilers anhand des Meergebnisses
; Frequenz am PIC darf 5 MHz nicht erreichen
; BIN4 enthlt Eingangsfrequenz in MHz
; >128MHz 62:1 - 1xxx xxxx - PS=5	1:64
; >64 MHz 32:1 - 11xx xxxx - PS=4	1:32
; >32 MHz 16:1 - 111x xxxx - PS=3	1:16
; >16 MHz  8:1 - 1111 xxxx - PS=2	1:8
; > 8 MHz  4:1 - 1111 1xxx - PS=1	1:4
; > 4 MHz  2:1 - 1111 11xx - PS=0	1:2
; < 4 MHz  1:1 -             PS=8	1:1

FSetPS
	movfw	BIN4
	movwf	MHz
	; falls die RA4-Eingangsfrequenz unter 4MHz liegt, 
	; wird kein prescaler verwendet
	movlw	0x08		; kein vorteiler
	movwf	PreSC
	movfw	MHz
	andlw	B'11111100'
	bz	PSNo		; falls BIN4 < 4 dann kein Vorteiler	
	movlw	0x06
	movwf	PreSC
PSDec
	decf	PreSC, f
	rlf	MHz, f
	btfss	STATUS, C
	goto	PSDec
	goto	PSEnd
PSNo
	; falls 64:1 Vorteiler, dann wenigstens 2:1 wegen asymmetrischem Rechtecksignal	
	btfss	Mode, 1		; Bei Mode 0 oder 1 berspringen
	goto	PSEnd
	movlw	0x00		; Vorteiler 2:1
	movwf	PreSC
PSEnd	
	movfw	BIN4		; MHz wider reparieren
	movwf	MHz
	return



;***********************************************************************************
; Pulslngenmessung fr 1 Hz .. 10 kHz
;
; zaehlt  mit 0,2us Takt ueber 4 Perioden -> 0.05 us Aufloesung
;   1 Hz = 5e6 /      1 x 4 =20 000 000 = 01 31 2D 00
;  10 Hz = 5e6 /     10 x 4 = 2 000 000 = 00 1E 84 80
; 100 Hz = 5e6 /    100 x 4 =   200 000 = 00 03 0D 40
; 200 Hz = 5e6 /    200 x 4 =   100 000 = 00 01 86 A0
; 250 Hz = 5e6 /    250 x 4 =    80 000 = 00 01 38 80
; 300 Hz = 5e6 /    300 x 4 =    66 666 = 00 01 04 6A
; 350 Hz = 5e6 /    350 x 4 =    57 142 = 00 00 DF 36

;  1 kHz = 5e6 /  1 000 x 4 =    20 000 = 00 4E 20 ->  1 000 Subtraktionen + 1 000 Subtraktionen
;  2 kHz = 5e6 /  2 000 x 4 =    10 000 = 00 27 10	Fehler: 0,20 Hz
;  3 kHz = 5e6 /  3 000 x 4 =     6 666 = 00 1A 0A	Fehler: 0,45 Hz
;  4 kHz = 5e6 /  4 000 x 4 =     5 000 = 00 13 88	Fehler: 0,80 Hz
;  5 kHz = 5e6 /  5 000 x 4 =     4 000 = 00 0F A0	Fehler: 1,25 Hz
;  6 kHz = 5e6 /  6 000 x 4 =     3 333 = 00 0D 05	Fehler: 1,80 Hzz
;  6,3kHz= 5e6 /  6 300 x 4 =     3 174 = 00 0C 66	Fehler: 2,00 Hz == oberhalb 6,3 kH ist Frequenzzaehler genauer
;  7 kHz = 5e6 /  7 000 x 4 =     2 857 = 00 0B 29	Fehler: 2,45 Hz !!
;  8 kHz = 5e6 /  8 000 x 4 =     2 500 = 00 09 C4	Fehler: 3,20 Hz !!
;  9 kHz = 5e6 /  9 000 x 4 =     2 222 = 00 08 AE	Fehler: 4,05 Hz !!
; 10 kHz = 5e6 / 10 000 x 4 =     2 000 = 00 07 D0 -> 10 000 Subtraktionen + 1 000 Subtraktionen = ca 500 000 cyklen = 0.1 s
;
; Ergebnis  nach xw und f
;
;***********************************************************************************
PMessung
;	call	LcdZeile1
;	movlw	'M'
;	movwf	LcdDaten
;	call	OutLcdDaten
	
	; Timer1 auf 0,2s-Takt einstellen fr Messung
	; interner Takt, Vorteiler 1:1
	; 20MHz / 4  = 5 MHz -> 0.2 us
	movlw	B'00000001'	; internen Takt zhlen, Vorteiler 1:1
	movwf	T1CON
	movlw	B'00000110'	; Capture mode, jede 4. steigende Flanke
	movwf	CCP1CON

	; es wird 2. ge-captured
	; berlufe des Timer1 zwischen den beiden Capture-Punkten werden in TMR1Loop mitgezhlt
	; Messzeit ist die Differenz beider Capture-Werte

	; vorbereiten fr den Beginn der Periode (1. Capture)
	clrf	TMR1L
	clrf	TMR1H
	clrf	TMR1LoopL
	clrf	TMR1LoopH
	bcf	PIR1,CCP1IF
	bcf	PIR1,TMR1IF

	; warten auf den Start
	; u.U. kann hier schon der Timer1 bergelaufen sein
	; das ist aber egal, es sei denn der berlauf war nach dem Auslsen von CCP
PerStart
	btfss	Taste			; Abbruch der Pulsweitenmessung
	goto	PerAbbruch		; Abbruch durch taste

	movfw	PIR1
	movwf	PIR1Puffer
	bcf	PIR1,TMR1IF		; bereit fr nchsten berlauf
	bcf	PIR1,CCP1IF
	; CCP1 und TMR1		-> falls CCPR1H=0 dann bertrag
	; nur CCP1		-> Start
	; nur TMR1		-> TMR1IF lschen und Loop
	; nichts		-> Loop
	btfsc	PIR1Puffer,TMR1IF
	goto	Startberlauf
	btfsc	PIR1Puffer,CCP1IF
	goto	MessenS			; nur CCP1, normales Ende
	goto	PerStart		; nicht passiert -> Loop

Startberlauf
	btfss	PIR1Puffer, CCP1IF
	goto	PerStart
	; was kam zuerst? ccp oder TMR1 ?
	; falls CCP zuerst kam, dann einen ebertrag zaehlen
	;movfw	CCPR1H			; 1-TMR1IF kam nach CCP -> bertrag;  0-TMR1IF kam vor CCP -> ignorieren
	btfss	CCPR1H,7		; 1-TMR1IF kam nach CCP -> bertrag;  0-TMR1IF kam vor CCP -> ignorieren
	;skpnz
	goto	MessenS			; TMR1IF kam vorher
	incf	TMR1LoopL,f		; bertrag, da CCP1H=0
	skpnz
	incf	TMR1LoopH,f

MessenS
	; CCP hat ausgelst, es kann los gehen
	; Mewert retten
	;bcf	PIR1,CCP1IF
	movfw	CCPR1L
	movwf	StartL
	movfw	CCPR1H
	movwf	StartH

	;warten auf das Ende
PerEnde
	movfw	PIR1
	movwf	PIR1Puffer
	; CCP1 und TMR1		-> unklar, falls CCPR1H,7=0 dann bertrag
	; nur CCP1		-> Ende
	; nur TMR1		-> bertrag und Loop
	; nichts		-> Loop

	btfsc	PIR1Puffer,TMR1IF
	goto	Uebertragtest
	btfsc	PIR1Puffer,CCP1IF
	goto	MessenE			; nur CCP1, normales Ende
	goto	PerEnde			; nicht passiert -> Loop

Uebertragtest
	; Timer 1 bergelaufen
	btfss	Taste			; Abbruch der Pulsweitenmessung
	goto	PerAbbruch		; Abbruch durch taste

	btfsc	PIR1Puffer,CCP1IF
	goto	PerBeide		; CCP und TMR1
	; nur TMR1
	bcf	PIR1,TMR1IF
	incf	TMR1LoopL,f
	skpnz
	incf	TMR1LoopH,f
	; dauert Messung zu lange ?
	btfsc	TMR1LoopH,1		; max. 512 Zyklen ~ 0.6Hz
	goto	Ueberlauf
	btfsc	PpmMode
	goto	PerEnde			; normale Schleife mit bertrag im PpmMode
	btfsc	TMR1LoopL,6		; max. 64 Zyklen ~ 4.8Hz	
	goto	Ueberlauf
	goto	PerEnde			; normale Schleife mit bertrag im NormalMode


PerBeide
	; beide Flags sind gesezt
	movfw	CCPR1H
	btfss	STATUS,Z
	goto	MessenE			; nach CCP, kein berlauf	
	incf	TMR1LoopL,f		; berlauf vor CCP
	skpnz
	incf	TMR1LoopH,f

	; dauert Messung zu lange ?
	btfsc	TMR1LoopH,1		; max. 512 Zyklen ~ 0.6Hz
	goto	Ueberlauf
	btfsc	PpmMode
	goto	MessenE	
	btfsc	TMR1LoopL,6		; max. 64 Zyklen ~ 4.8Hz	
	goto	Ueberlauf
MessenE
	; Mewert retten
	movfw	CCPR1L
	movwf	PeriodeL
	movfw	CCPR1H
	movwf	PeriodeH
	bcf	PIR1,CCP1IF

	; Differenz aus Ende und Anfang berechnen
	movfw	TMR1LoopH
	movwf	f3
	movfw	TMR1LoopL
	movwf	f2
	movfw	PeriodeH
	movwf	f1
	movfw	PeriodeL
	movwf	f0

	clrf	xw3
	clrf	xw2
	movfw	StartH
	movwf	xw1
	movfw	StartL
	movwf	xw0

	call	Sub32		; 32 bit f:=f-xw   calc=xw cnt=f
	; Ergebnis nach xw laden
	movfw	f0
	movwf	xw0
	movfw	f1
	movwf	xw1
	movfw	f2
	movwf	xw2
	movfw	f3
	movwf	xw3

	; falls das Ergebnis < 0x00 08 00 ist, dann liegen ber 9,765 kHz an
	; die Berechnung wre sehr langsam
	; es ist besser, in den Zhlermode zu gehen
	movfw	f2
	iorwf	f3, w
	btfss	STATUS, Z
	goto	NichtZuHoch	; unter 30 Hz -> ok
	movfw	f1
	andlw	0xF8
	btfss	STATUS, Z
	goto	NichtZuHoch	; unter 10 kHz -> ok
	bcf	Pulsmode	; FAIL; Frequenz zu hoch
	return
NichtZuHoch
	return
PerAbbruch
	bsf	Buttonpressed
Ueberlauf
	clrf	f3
	clrf	f2
	clrf	f1
	clrf	f0
	bcf	Pulsmode
	return


;**************************************************************
; Neuberechnung des Anzeigewerts
;
; input:  xw = Messwert in xw
; output: f  = Frequenz x 1 000		(Pulsmode3)
;         f  = Frequenz x 1 000 000	(Pulsmode6)
;         i  = Messwert
;**************************************************************
PBerechnung
;	call	LcdZeile1
;	movlw	'R'
;	movwf	LcdDaten
;	call	OutLcdDaten

	;debug  einfach den halben zaehlwert anzeigen, Periode in 0.1us Aufloesung
;	movfw	xw0
;	movwf	f0
;	movfw	xw1
;	movwf	f1
;	movfw	xw2
;	movwf	f2
;	movfw	xw3
;	movwf	f3
;	bcf	STATUS,C
;	rrf	f3, f
;	rrf	f2, f
;	rrf	f1, f
;	rrf	f0, f
;	return


	; messwert steht in xw
	; 20 000 000 durch Mewert teilen
	; 20 000 000d = 0x01 31 2D 00

	; 20 000 000 -> f 
	movlw	0x01
	movwf	f3
	movlw	0x31
	movwf	f2
	movlw	0x2D
	movwf	f1
	movlw	0x00
	movwf	f0

; falls der Zhlwert in xw > 0x41 89 37
; dann knnen die spteren Multiplikationen des Restes mit 1000 grer als 32 bit werden.
; deshalb wird bei Zhlwerten ab  0x40 00 00 der Zhlwert und die 20000000 geachtelt
	movfw	xw2
	andlw	0xC0
	iorwf	xw3, w
	bz	Ber100		; <0x00 40 00 00
	;>0x00 40 00 00
	bcf	STATUS,C
	rrf	f3, f
	rrf	f2, f
	rrf	f1, f
	rrf	f0, f
	bcf	STATUS,C
	rrf	f3, f
	rrf	f2, f
	rrf	f1, f
	rrf	f0, f
	bcf	STATUS,C
	rrf	f3, f
	rrf	f2, f
	rrf	f1, f
	rrf	f0, f
	bcf	STATUS,C
	rrf	xw3, f
	rrf	xw2, f
	rrf	xw1, f
	rrf	xw0, f
	bcf	STATUS,C
	rrf	xw3, f
	rrf	xw2, f
	rrf	xw1, f
	rrf	xw0, f
	bcf	STATUS,C
	rrf	xw3, f
	rrf	xw2, f
	rrf	xw1, f
	rrf	xw0, f

Ber100
	; nun durch den Mewert dividieren
	call	Div32		; f:= f / xw   ; Int-Ergebnis in f,  Rest in r0..r3

	; Messwert in xw mu in i gepuffert werden, da Mal1000F32 xw verndert.
	movfw	xw0
	movwf	i0
	movfw	xw1
	movwf	i1
	movfw	xw2
	movwf	i2
	movfw	xw3
	movwf	i3		; messwert -> i

	; wieviele Nachkommastellen?
	bcf	Pulsmode3
	bcf	Pulsmode6
	; wenn h<64 (unter 64 Hz) dann mit 6 Nachkommastellen, ansonsten mit 3 Nachkommastellen
	movfw	f0
	andlw	0xC0
	iorwf	f1, w
	iorwf	f2, w
	iorwf	f3, w
	bnz	Per3	; >63 -> 3 Nachkommastellen
	bsf	Pulsmode6
	goto	PerE
Per3
	bsf	Pulsmode3
PerE

	; ganzzahligen Teil in h0..h3 x 1000
	call	Mal1000f32	; f:=f*1000
	movfw	f0
	movwf	h0
	movfw	f1
	movwf	h1
	movfw	f2
	movwf	h2
	movfw	f3
	movwf	h3	; f -> h

	; rest nach f
	movfw	r0
	movwf	f0
	movfw	r1
	movwf	f1
	movfw	r2
	movwf	f2
	movfw	r3
	movwf	f3		; r -> f
	call	Mal1000f32	; f:=f*1000
	; falls rest>4294967d = 0x418937 dann schlgt das fehl
	; das ist der Fall bei f<4,00 Hz

	; den Messwert xw aus i zurckholen
	movfw	i0
	movwf	xw0
	movfw	i1
	movwf	xw1
	movfw	i2
	movwf	xw2
	movfw	i3
	movwf	xw3		; i-> xw

	call	Div32		; f:= f / xw   ; Rest in r0..r3

	; Zahl zusammenbasteln
	movfw	h0
	movwf	xw0
	movfw	h1
	movwf	xw1
	movfw	h2
	movwf	xw2
	movfw	h3
	movwf	xw3
	call	Add32		; f:=f+xw

	btfss	Pulsmode6
	return			; mit 3 Nachkommastellen fertig

	; weiter fuer 6 Nachkommastellen
	call	Mal1000f32	; f:=f*1000 bisheriges Ergebnis * 1000
	movfw	f0
	movwf	h0
	movfw	f1
	movwf	h1
	movfw	f2
	movwf	h2
	movfw	f3
	movwf	h3		; Ergebnis*1000 -> h


	; Rest fr die Nachkommastellen
	movfw	r0
	movwf	f0
	movfw	r1
	movwf	f1
	movfw	r2
	movwf	f2
	movfw	r3
	movwf	f3		; r -> f
	call	Mal1000f32	; f:=f*1000

	; xw aus i zurckholen
	movfw	i0
	movwf	xw0
	movfw	i1
	movwf	xw1
	movfw	i2
	movwf	xw2
	movfw	i3
	movwf	xw3

	call	Div32		; f:= f / xw   ; Rest in r0..r3

	; Zahl zusammenbasteln
	movfw	h0
	movwf	xw0
	movfw	h1
	movwf	xw1
	movfw	h2
	movwf	xw2
	movfw	h3
	movwf	xw3
	call	Add32		; f:=f+xw

	return			; mit 6 Nachkommastellen fertig


;**************************************************************
Display
	call	LcdZeile1
	;Binr zu ASCII Wandlung mit Ausgabe
	btfsc	Pulsmode
	goto	Display2
	bcf	Pulsmode3
	bcf	Pulsmode6
Display2
	call	OutDez32	; ' x xxx xxx xxx'
	call	OutHz		; ' Hz'
	return


;***********************************************************************************
;***********************************************************************************
; mit der Eingangsled blinken
InputFlash
	; alle relais aus
	bsf	PORTA, 0
	bsf	PORTA, 1
	bsf	PORTA, 2
	call	Wait200ms

	; relais wieder an
	movfw	Mode
	btfsc	STATUS, Z	; =0?
	bcf	PORTA, 0
	btfsc	Mode,0		; =1?
	bcf	PORTA, 1
	btfsc	Mode,1		; =2?
	bcf	PORTA, 2
	call	Wait200ms
	return



; Eingang zuschalten falls gendert
; Relais werden mit LOW eingeschaltet
Inputselect
	btfss	Inputchanged
	return
	; alle relais aus
	bsf	PORTA, 0
	bsf	PORTA, 1
	bsf	PORTA, 2
	movfw	Mode
	btfsc	STATUS, Z	; =0?
	bcf	PORTA, 0
	btfsc	Mode,0		; =1?
	bcf	PORTA, 1
	btfsc	Mode,1		; =2?
	bcf	PORTA, 2
	bcf	Inputchanged
	call	Wait100ms	; 100ms fr Relais
	return



;***********************************************************************************
;***********************************************************************************
; Taster auswerten:
;Taster ist low-aktiv
CheckTaster
	bcf	Buttonpressed
	btfsc	Taste
	return			; nicht gedrueckt
	;taste gedrueckt - mal sehen wie lange der gedrueck wird
	bcf	PpmMode
	bcf	PpmMode2
	clrf	TastCounter
TasteLoop	
	call	Wait50ms
	incf	TastCounter, f
	bz	TasteUeberlauf	; 256 x 50ms = 12.8 s
	; 1 s = PpmMode
	movfw	TastCounter
	xorlw	0x14		; 20d,  20 x 50ms = 1s
	bnz	Taste2s
	; 1 Sekunde vorbei
	call	LcdZeile1_1
	call	OutStern
	call	InputFlash	; 2 x blinken
	call	InputFlash
	bsf	PpmMode
Taste2s
	; 2 s = PpmMode2
	movfw	TastCounter
	xorlw	0x28		; 40d,  40 x 50ms = 2s
	bnz	Taste3s
	; 2 Sekunde vorbei
	call	LcdZeile1_1
	call	OutSharp
	call	InputFlash	; 4 x blinken
	call	InputFlash
	call	InputFlash
	call	InputFlash
	bsf	PpmMode2
	bsf	PpmMode		; ist eigentlich schon gesetzt
Taste3s
	btfss	Taste
	goto	TasteLoop
	; wie lange gedrueckt?

TasteUeberlauf
	btfsc	PpmMode
	return			; > 1s
	btfsc	PpmMode2
	return			; > 2s

	; < 1s - Inputwechsel
	call	LcdZeile1
	call	OutSpace
	bsf	Inputchanged
	incf	Mode, f		; Modenummer erhhen
	btfss	Mode, 0		; >2 ?
	goto	TastModeEnd	; nein
	btfss	Mode, 1
	goto	TastModeEnd	; nein
	clrf	Mode		; ja also =0
TastModeEnd
	call	LcdShowInput	; Signaleingang anzeigen
	call	Inputselect	; Relais ansteuern
	call	SaveMode	; -> EEPROM
	return
	



;*********************************************************************
;*** INCLUDE *********************************************************
;*********************************************************************
#include	uhr.asm
#include	math.asm
#include	lcd.asm
#include	eeprom.asm


	end		
