
;**************************************************************
;**************************************************************
; die Uhr fr die Quarz-Kalibrierung
;
;**************************************************************
Uhr
	movlw   Ini_con     	; Interupt disable
	movwf   INTCON   

	clrf	Flags
	clrf	UFlags

	; Zeit 0 einstellen
	call	Zeit_Null

	; 19531.25 Hz-Timer0-Interupt; 20 MHz / 4 / 256 = 19531.25 Hz (51.2us)
	bsf     STATUS, RP0	; auf Bank 1 umschalten
	movlw	B'00001000'	; pull-up, internen Takt zhlen, kein Vorteiler
	movwf	OPTION_REG
	bcf     STATUS, RP0	; auf Bank 0 zurckschalten
	clrf	TMR0
	bsf	INTCON, T0IE	; Timer0 interupt erlauben
	bsf	INTCON, GIE	; Interupt erlauben

	clrf	DCF_onoff

Uhrloop
	btfsc	DCF_newpulse
	call	DCF_OutPulse	; DCF-Sekundenpuls

	call	Wait20ms

	btfss	Taste
	call	Zeit_Null	; wenn Taste gedrckt, dann Zeit=0

	btfsc	Taste
	clrf	DCF_onoff	; DCF-Aktivierungszaehler loeschen

	btfsc	DCF_todo
	call	DCF_Calc	; mintliche DCF-Berechnung

	btfsc	DCF_active
	goto	Uhrloop		; keine Zeitausgabe im DCF-Mode

	btfsc	neueZeit
	call	DisplayZeit

	goto	Uhrloop		; eine Endlosschleife



;**************************************************************
; Ausgabe der quarzgemessenen Uhrzeit
DisplayZeit
	; Kursor home
	movlw	B'10000000'	; 1. Zeile
	call	OutLcdControl
	movfw	ZH
	iorlw	'0'		;wandeln in ASCCI
	call	OutLcdDaten
	movfw	EH
	iorlw	'0'		;wandeln in ASCCI
	call	OutLcdDaten
	movlw	':'		; Lcke
	call	OutLcdDaten
	movfw	ZM
	iorlw	'0'		;wandeln in ASCCI
	call	OutLcdDaten
	movfw	EM
	iorlw	'0'		;wandeln in ASCCI
	call	OutLcdDaten
	movlw	':'		; Lcke
	call	OutLcdDaten
	movfw	ZS
	iorlw	'0'		;wandeln in ASCCI
	call	OutLcdDaten
	movfw	ES
	iorlw	'0'		;wandeln in ASCCI
	call	OutLcdDaten
	bcf	neueZeit
	return


;**************************************************************
; 1. Test ob DCF aktiviert werden soll
; 2. Uhrzeit auf 00:00:00 zurcksetzen
Zeit_Null
	; 256 x 20 ms = 5 s taste druecken um DCF zu aktivieren
	incf	DCF_onoff, f
	bnz	Zeit_Null_2
	call	DCF_Init	; DCF an

Zeit_Null_2
	clrf	TMR0
	bcf	neueZeit
	movlw	0x4C+1
	movwf	countH		; mu auf 0x4B+1 stehen, sonst funktioniert Abbruch bei dec nicht
	movlw	0x4B
	movwf	countL
	bsf	neueZeit
	clrf	ES
	clrf	ZS
	clrf	EM
	clrf	ZM
	clrf	EH
	clrf	ZH
	movlw	4
	movwf	Digit		; 4Hz -> 1 Hz Teiler
	return

;**************************************************************


;**************************************************************
;
; DCF77-Routinen
;
;**************************************************************
; 20.000.000 x 60 = 1.200.000.000 = 2x2x2x2x2x2x2x2x2x2 x 1171875
;                                 = 2^10 * 5x5x5x5x5x5x5x5 x3
;                                 = 2^10 * 5^8 * 3
;                                 = 0x47 86 8C 00
;                                 = 4 * 8 * 0x02 3C 34 60



;**************************************************************
; DCF77-Mode aktivieren
DCF_Init
	;'DCF77 wait '
	call	OutDCF
	movlw	'w'
	movwf	LcdDaten
	call	OutLcdDaten
	movlw	'a'
	movwf	LcdDaten
	call	OutLcdDaten
	movlw	'i'
	movwf	LcdDaten
	call	OutLcdDaten
	movlw	't'
	movwf	LcdDaten
	call	OutLcdDaten
	movlw	' '
	movwf	LcdDaten
	call	OutLcdDaten
	
	; Timer1
	; interner Takt, Vorteiler 8:1
	; 20MHz / 4 / 8 = 625 kHz -> 1.6 us
	movlw	B'00110001'	; internen Takt zhlen, Vorteiler 8:1
	movwf	T1CON
	;movlw	B'00000100'	; Capture mode, jede fallende Flanke
	movlw	B'00000101'	; Capture mode, jede steigende Flanke
	movwf	CCP1CON
	; ueberlauf nach 0.1048... s
	; 1 Minute = 0x023C Ueberlaefe + 0x3460 = 572d Ueberlaufe + 13408d = 0x02 3C 34 60

	; int fuer CCP1 und Timer1 erlauben
	bcf	PIR1, CCP1IF	; Interupt-Flag lschen
	bcf	PIR1, TMR1IF	; Interupt-Flag lschen
	bsf     STATUS, RP0	; Bank 1
	bsf	PIE1, CCP1IE
	bsf	PIE1, TMR1IE
	bcf     STATUS, RP0	; Bank 1
	bsf	INTCON, PEIE

	; TTL-Eingang einschalten
	bcf	PORTA, 1

	; 0x7270 = 1,5 Sekunden
	movlw	0x72+1
	movwf	DCFcountH
	movlw	0x70
	movwf	DCFcountL

	clrf	xw0
	clrf	xw1	
	clrf	xw2
	clrf	xw3

	clrf	f0
	clrf	f1
	clrf	f2
	clrf	f3
	
	clrf	d0	; summe ueber 8 minuten
	clrf	d1
	clrf	d2
	clrf	d3
	
	clrf	h0
	clrf	h1
	clrf	s0
	clrf	s1
	clrf	UFlags
	clrf	DCF_onoff
	clrf	DCF_Quality
	clrf	DCF_pulscount
	;clrf	DCF_Minuten
	movlw	0xFF
	movwf	DCF_Minuten

	bsf	DCF_active

	return


;**************************************************************
; DCF77-Mode abschalten
DCF_Off
	; interrups CCP1 und TMR1 aus
	bcf	INTCON, PEIE
	bsf     STATUS, RP0	; Bank 1
	bcf	PIE1, CCP1IE
	bcf	PIE1, TMR1IE
	bcf     STATUS, RP0	; Bank 1

	clrf	T1CON		; Timer1 aus
	clrf	CCP1CON		; CCP1 aus

	bcf	DCF_active
	clrf	UFlags

	; TTL-Eingang abschalten
	bsf	PORTA, 1

	;'DCF77 fail '
	call	OutDCF
	movlw	'f'
	movwf	LcdDaten
	call	OutLcdDaten
	movlw	'a'
	movwf	LcdDaten
	call	OutLcdDaten
	movlw	'i'
	movwf	LcdDaten
	call	OutLcdDaten
	movlw	'l'
	movwf	LcdDaten
	call	OutLcdDaten
	movlw	' '
	movwf	LcdDaten
	call	OutLcdDaten
	return


;**************************************************************
; Ausgabe der Sekundenpulsnummer
; input: DCF_newpulse = 1
;        DCF_pulscount= Sekundenpulsnummer
; output:DCF_newpulse = 0
DCF_OutPulse
	bcf	DCF_newpulse
	call	LcdZeile1_10
	movlw	'P'
	movwf	LcdDaten
	call	OutLcdDaten
	movfw	DCF_pulscount
	movwf	f0
	clrf	f1
	clrf	f2
	clrf	f3
	call	Hex2Dez8

	clrf	Fehler
	movfw	HdH
	Call	Leading0	; 1 111 111 1xx 
	movfw	HdZ
	Call	Leading0	; 1 111 111 11x 
	movfw	HdE
	Call	Bcd4Bit		; 1 111 111 111 
	return

;**************************************************************
; berechnet aus dem Abstand zweier Minutenpulse den Quarzfehler
; input : s  Zeitpunkt des vorigen Pulses (16 bit)
;         h  Zeitpunkt des jetzigen Pulses (32 bit)
DCF_Calc
	btfss	DCF_ok
	goto	DCF_Calc_end

	; endwert -> f
	movfw	h0
	movwf	f0
	movfw	h1
	movwf	f1
	movfw	h2
	movwf	f2
	movfw	h3
	movwf	f3
	
	; startwert s -> xw
	movfw	s0
	movwf	xw0
	movfw	s1
	movwf	xw1
	clrf	xw2
	clrf	xw3

	call	Sub32		; f = Endzeit - Startzeit = 1 Minute

	; sollwert = 20MHz /4 /8 *60 = 37500000 = 0x02 3C 34 60
	; sollwert = 0x02 3C 34 60 -> xw
	movlw	0x60
	movwf	xw0
	movlw	0x34
	movwf	xw1
	movlw	0x3C
	movwf	xw2
	movlw	0x02
	movwf	xw3
	bcf	DCF_negativ
	call	Sub32		; f= Fehler in 1,6-us-increments (0,0266 ppm)
	skpnc
	bsf	DCF_negativ

	; nun das aufaddieren zum 8-Minutenfehler
	btfsc	DCF_Minuten,7	; Minute = -1 ?
	goto	DCF_Calc_Justierminute

	; Minutenfehler f -> xw
	movfw	f0
	movwf	xw0
	movfw	f1
	movwf	xw1
	movfw	f2
	movwf	xw2
	movfw	f3
	movwf	xw3

	; 8-Minutensumme d -> f
	movfw	d0
	movwf	f0
	movfw	d1
	movwf	f1
	movfw	d2
	movwf	f2
	movfw	d3
	movwf	f3

	call	Add32	; addiere alle Minutenfehler

	; 8-Minutensumme f -> d
	movfw	f0
	movwf	d0
	movfw	f1
	movwf	d1
	movfw	f2
	movwf	d2
	movfw	f3
	movwf	d3

	; Minutenfehler xw -> f
	movfw	xw0
	movwf	f0
	movfw	xw1
	movwf	f1
	movfw	xw2
	movwf	f2
	movfw	xw3
	movwf	f3
DCF_Calc_Justierminute

	call	DisplayFehler	; Minutenfehler ausgeben

	incf	DCF_Minuten,f
	btfss	DCF_Minuten, 3	; >= 8 minuten?
	goto	DCF_Calc_Minu

	; alle 8 Minuten
	; 8-Minutensumme d -> f
	movfw	d0
	movwf	f0
	movfw	d1
	movwf	f1
	movfw	d2
	movwf	f2
	movfw	d3
	movwf	f3

	call	Display8Minuten	; 8 Minutenfehler ausgeben

	clrf	d0
	clrf	d1
	clrf	d2
	clrf	d3
	;clrf	DCF_Minuten
	movlw	0xFF
	movwf	DCF_Minuten
	goto	DCF_Calc_end

	; Ausgabe der Warteminuten
DCF_Calc_Minu
	call	LcdZeile1
	movlw	'W'
	movwf	LcdDaten
	call	OutLcdDaten
	movlw	'a'
	movwf	LcdDaten
	call	OutLcdDaten
	movlw	'i'
	movwf	LcdDaten
	call	OutLcdDaten
	movlw	't'
	movwf	LcdDaten
	call	OutLcdDaten
	movlw	' '
	movwf	LcdDaten
	call	OutLcdDaten
	movfw	DCF_Minuten
	sublw	8
	iorlw	'0'
	movwf	LcdDaten
	call	OutLcdDaten
	movlw	'm'
	movwf	LcdDaten
	call	OutLcdDaten
	movlw	'i'
	movwf	LcdDaten
	call	OutLcdDaten
	movlw	'n'
	movwf	LcdDaten
	call	OutLcdDaten


DCF_Calc_end	
	movfw	h0		; Endwert als neuen Startwert festlegen
	movwf	s0
	movfw	h1
	movwf	s1

	bcf	DCF_todo	; Berechnung erledigt
	bsf	DCF_ok		; DCF-Pulse empfangen, gueltiger Startwert in xw
	return



;**************************************************************
; Anzeige des gemessenen Quarzfehlers nach 8 Minuten
; input f Zahlenwert in 0,02666 ppm Aufloesung
Display8Minuten
	call	LcdZeile1
	movlw	'+'
	btfsc	f3,7
	movlw	'-'
	movwf	LcdDaten
	call	OutLcdDaten

	btfss	f3,7
	goto	Display8positiv
	; negativ
	movfw	f0
	movwf	xw0
	movfw	f1
	movwf	xw1
	movfw	f2
	movwf	xw2
	movfw	f3
	movwf	xw3

	clrf	f0
	clrf	f1
	clrf	f2
	clrf	f3

	call	Sub32	; f = 0 - f

	; durch 30 dividieren
	movlw	0x1E	; 30
	movwf	xw0
	clrf	xw1
	clrf	xw2
	clrf	xw3
	call	Div32	; f=f/xw,  rest->g

Display8positiv
	bsf	DCF_dppm
	call	OutDez16L0	; ' xx xxx'
	bcf	DCF_dppm
	movlw	'p'
	movwf	LcdDaten
	call	OutLcdDaten
	movlw	'p'
	movwf	LcdDaten
	call	OutLcdDaten
	movlw	'm'
	movwf	LcdDaten
	call	OutLcdDaten

	return

;**************************************************************
; Anzeige des gemessenen 1 Minuten-Quarzfehlers in 0,02666 ppm Aufloesung
; input C-Flag : negativ
;       f Zahlenwert
DisplayFehler
	btfss	f3,7
	goto	DisplayFehlerPos
	; negativ
	movfw	f0
	movwf	xw0
	movfw	f1
	movwf	xw1
	movfw	f2
	movwf	xw2
	movfw	f3
	movwf	xw3

	clrf	f0
	clrf	f1
	clrf	f2
	clrf	f3

	call	Sub32	; f = 0 - f

DisplayFehlerPos
	call	LcdZeile2
	movlw	'+'
	btfsc	DCF_negativ
	movlw	'-'
	movwf	LcdDaten
	call	OutLcdDaten
	call	OutDez32	; ' x xxx xxx xxx'

	movlw	' '
	movwf	LcdDaten
	call	OutLcdDaten

	; wenigstens 2 mal 58 saubere Pulse?
	movfw	DCF_Quality
	andlw	0xFE
	bz	DisplayFehlerLow
	movlw	'#'
	movwf	LcdDaten
	call	OutLcdDaten
	return
DisplayFehlerLow
	movlw	' '
	movwf	LcdDaten
	call	OutLcdDaten
	return

;**************************************************************
; 'DCF77 '
OutDCF
	call	LcdZeile2
	movlw	'D'
	movwf	LcdDaten
	call	OutLcdDaten
	movlw	'C'
	movwf	LcdDaten
	call	OutLcdDaten
	movlw	'F'
	movwf	LcdDaten
	call	OutLcdDaten
	movlw	'7'
	movwf	LcdDaten
	call	OutLcdDaten
	movlw	'7'
	movwf	LcdDaten
	call	OutLcdDaten
	movlw	' '
	movwf	LcdDaten
	call	OutLcdDaten
	return

;**************************************************************